@@ -218,6 +218,7 @@ bool reload_acl_and_cache(THD *thd, unsigned long long options,
               thd->handler_tables_hash.records ||
               thd->ull_hash.records ||
               thd->global_read_lock.is_acquired() ||
+              thd->mdl_backup_lock ||
               thd->current_backup_stage != BACKUP_FINISHED
               );
 
@@ -530,7 +531,14 @@ bool flush_tables_with_read_lock(THD *thd, TABLE_LIST *all_tables)
   if (thd->current_backup_stage != BACKUP_FINISHED)
   {
     my_error(ER_BACKUP_LOCK_IS_ACTIVE, MYF(0));
-    return true;
+    goto error;
+  }
+
+  /* Should not flush tables while BACKUP LOCK is active */
+  if (thd->mdl_backup_lock)
+  {
+    my_error(ER_LOCK_OR_ACTIVE_TRANSACTION, MYF(0));
+    goto error;
   }
 
   if (thd->lex->type & REFRESH_READ_LOCK)
