@@ -40,7 +40,6 @@ static void init_service_thd(THD* thd, char* thread_stack)
   thd->prior_thr_create_utime= thd->start_utime= microsecond_interval_timer();
   thd->set_command(COM_SLEEP);
   thd->reset_for_next_command(true);
-  server_threads.insert(thd); // as wsrep_innobase_kill_one_trx() uses find_thread_by_id()
 }
 
 Wsrep_storage_service*
@@ -80,7 +79,6 @@ void Wsrep_server_service::release_storage_service(
     static_cast<Wsrep_storage_service*>(storage_service);
   THD* thd= ss->m_thd;
   wsrep_reset_threadvars(thd);
-  server_threads.erase(thd);
   delete ss;
   delete thd;
 }
@@ -94,8 +92,7 @@ wsrep_create_streaming_applier(THD *orig_thd, const char *ctx)
      streaming transaction is BF aborted and streaming applier
      is created from BF aborter context. */
   Wsrep_threadvars saved_threadvars(wsrep_save_threadvars());
-  if (saved_threadvars.cur_thd)
-    wsrep_reset_threadvars(saved_threadvars.cur_thd);
+  wsrep_reset_threadvars(saved_threadvars.cur_thd);
   THD *thd= 0;
   Wsrep_applier_service *ret= 0;
   if (!wsrep_create_threadvars() &&
@@ -112,8 +109,7 @@ wsrep_create_streaming_applier(THD *orig_thd, const char *ctx)
   }
   /* Restore original thread local storage state before returning. */
   wsrep_restore_threadvars(saved_threadvars);
-  if (saved_threadvars.cur_thd)
-    wsrep_store_threadvars(saved_threadvars.cur_thd);
+  wsrep_store_threadvars(saved_threadvars.cur_thd);
   return ret;
 }
 
@@ -142,7 +138,6 @@ void Wsrep_server_service::release_high_priority_service(wsrep::high_priority_se
   THD* thd= hps->m_thd;
   delete hps;
   wsrep_store_threadvars(thd);
-  server_threads.erase(thd);
   delete thd;
   wsrep_delete_threadvars();
 }
@@ -158,7 +153,7 @@ void Wsrep_server_service::bootstrap()
   wsrep::log_info()
     << "Bootstrapping a new cluster, setting initial position to "
     << wsrep::gtid::undefined();
-  wsrep_set_SE_checkpoint(wsrep::gtid::undefined());
+  wsrep_set_SE_checkpoint(wsrep::gtid::undefined(), wsrep_gtid_server.undefined());
 }
 
 void Wsrep_server_service::log_message(enum wsrep::log::level level,
@@ -167,19 +162,16 @@ void Wsrep_server_service::log_message(enum wsrep::log::level level,
   switch (level)
   {
   case wsrep::log::debug:
-    WSREP_DEBUG("%s", message);
+    sql_print_information("debug: %s", message);
     break;
   case wsrep::log::info:
-    WSREP_INFO("%s", message);
+    sql_print_information("%s", message);
     break;
   case wsrep::log::warning:
-    WSREP_WARN("%s", message);
+    sql_print_warning("%s", message);
     break;
   case wsrep::log::error:
-    WSREP_ERROR("%s", message);
-    break;
-  case wsrep::log::unknown:
-    WSREP_UNKNOWN("%s", message);
+    sql_print_error("%s", message);
     break;
   }
 }
@@ -221,7 +213,7 @@ void Wsrep_server_service::log_view(
       if (prev_view.state_id().id() != view.state_id().id())
       {
         WSREP_DEBUG("New cluster UUID was generated, resetting position info");
-        wsrep_set_SE_checkpoint(wsrep::gtid::undefined());
+        wsrep_set_SE_checkpoint(wsrep::gtid::undefined(), wsrep_gtid_server.undefined());
         checkpoint_was_reset= true;
       }
 
@@ -256,7 +248,7 @@ void Wsrep_server_service::log_view(
             WSREP_WARN("Failed to commit transaction for store view");
           }
         }
-        applier->m_thd->release_transactional_locks();
+        applier->m_thd->mdl_context.release_transactional_locks();
       }
 
       /*
@@ -272,9 +264,9 @@ void Wsrep_server_service::log_view(
           Wsrep_server_state::instance().provider().last_committed_gtid().seqno();
       if (checkpoint_was_reset || last_committed != view.state_id().seqno())
       {
-          wsrep_set_SE_checkpoint(view.state_id());
+        wsrep_set_SE_checkpoint(view.state_id(), wsrep_gtid_server.gtid());
       }
-      DBUG_ASSERT(wsrep_get_SE_checkpoint().id() == view.state_id().id());
+      DBUG_ASSERT(wsrep_get_SE_checkpoint<wsrep::gtid>().id() == view.state_id().id());
     }
     else
     {
@@ -308,7 +300,7 @@ wsrep::view Wsrep_server_service::get_view(wsrep::client_service& c,
 
 wsrep::gtid Wsrep_server_service::get_position(wsrep::client_service&)
 {
-  return wsrep_get_SE_checkpoint();
+  return wsrep_get_SE_checkpoint<wsrep::gtid>();
 }
 
 void Wsrep_server_service::set_position(wsrep::client_service& c WSREP_UNUSED,
@@ -326,7 +318,7 @@ void Wsrep_server_service::set_position(wsrep::client_service& c WSREP_UNUSED,
     WSREP_WARN("Wait for gtid returned error %d while waiting for "
                "prior transactions to commit before setting position", err);
   }
-  wsrep_set_SE_checkpoint(gtid);
+  wsrep_set_SE_checkpoint(gtid, wsrep_gtid_server.gtid());
 }
 
 void Wsrep_server_service::log_state_change(
