@@ -768,7 +768,6 @@ bool JDBConn::Connect(PJPARM sop)
 {
 	bool		 err = false;
 	jint     rc;
-	jboolean jt = (trace(1));
 	PGLOBAL& g = m_G;
 
 	/*******************************************************************/
@@ -938,7 +937,7 @@ int JDBConn::Rewind(PCSZ sql)
 		if (gmID(m_G, fetchid, "Fetch", "(I)Z"))
 			return -1;
 
-		jboolean b = env->CallBooleanMethod(job, fetchid, 0);
+		(void) env->CallBooleanMethod(job, fetchid, 0);
 
 		rbuf = m_Rows;
 	} else if (ExecuteCommand(sql) != RC_FX)
@@ -1190,7 +1189,7 @@ int JDBConn::ExecuteUpdate(PCSZ sql)
 /***********************************************************************/
 int JDBConn::GetResultSize(PCSZ sql, PCOL colp)
 {
-	int rc, n = 0;
+	int rc;
 
 	if ((rc = ExecuteQuery(sql)) != RC_OK)
 		return -1;
@@ -1497,7 +1496,6 @@ bool JDBConn::SetParam(JDBCCOL *colp)
 		PCSZ     fnc = "Unknown";
 		uint     n;
 		short    len, tp;
-		int      crow = 0;
 		PQRYRES  qrp = cap->Qrp;
 		PCOLRES  crp;
 		jboolean rc = false;
