@@ -43,7 +43,6 @@ PageBulk::init()
 {
 	buf_block_t*	new_block;
 	page_t*		new_page;
-	ulint		new_page_no;
 
 	ut_ad(m_heap == NULL);
 	m_heap = mem_heap_create(1000);
@@ -61,12 +60,10 @@ PageBulk::init()
 		alloc_mtr.start();
 		m_index->set_modified(alloc_mtr);
 
-		ulint	n_reserved;
-		bool	success;
-		success = fsp_reserve_free_extents(&n_reserved,
-						   m_index->table->space,
-						   1, FSP_NORMAL, &alloc_mtr);
-		if (!success) {
+		uint32_t n_reserved;
+		if (!fsp_reserve_free_extents(&n_reserved,
+					      m_index->table->space,
+					      1, FSP_NORMAL, &alloc_mtr)) {
 			alloc_mtr.commit();
 			m_mtr.commit();
 			return(DB_OUT_OF_FILE_SPACE);
@@ -81,7 +78,7 @@ PageBulk::init()
 		alloc_mtr.commit();
 
 		new_page = buf_block_get_frame(new_block);
-		new_page_no = page_get_page_no(new_page);
+		m_page_no = new_block->page.id().page_no();
 
 		byte* index_id = my_assume_aligned<2>
 			(PAGE_HEADER + PAGE_INDEX_ID + new_page);
@@ -108,8 +105,7 @@ PageBulk::init()
 					  false, &m_mtr);
 
 		new_page = buf_block_get_frame(new_block);
-		new_page_no = page_get_page_no(new_page);
-		ut_ad(m_page_no == new_page_no);
+		ut_ad(new_block->page.id().page_no() == m_page_no);
 
 		ut_ad(page_dir_get_n_heap(new_page) == PAGE_HEAP_NO_USER_LOW);
 
@@ -125,7 +121,6 @@ PageBulk::init()
 
 	m_block = new_block;
 	m_page = new_page;
-	m_page_no = new_page_no;
 	m_cur_rec = page_get_infimum_rec(new_page);
 	ut_ad(m_is_comp == !!page_is_comp(new_page));
 	m_free_space = page_get_free_space_of_empty(m_is_comp);
@@ -854,6 +849,8 @@ PageBulk::latch()
 	m_mtr.start();
 	m_index->set_modified(m_mtr);
 
+	ut_ad(m_block->page.buf_fix_count());
+
 	/* In case the block is S-latched by page_cleaner. */
 	if (!buf_page_optimistic_get(RW_X_LATCH, m_block, m_modify_clock,
 				     __FILE__, __LINE__, &m_mtr)) {
@@ -872,6 +869,8 @@ PageBulk::latch()
 
 	buf_block_buf_fix_dec(m_block);
 
+	ut_ad(m_block->page.buf_fix_count());
+
 	ut_ad(m_cur_rec > m_page && m_cur_rec < m_heap_top);
 
 	return (m_err);
@@ -981,7 +980,7 @@ inline void BtrBulk::logFreeCheck()
 	if (log_sys.check_flush_or_checkpoint()) {
 		release();
 
-		log_free_check();
+		log_check_margins();
 
 		latch();
 	}
@@ -1111,13 +1110,9 @@ BtrBulk::insert(
 				goto func_exit;
 			}
 
-			/* Wake up page cleaner to flush dirty pages. */
 			srv_inc_activity_count();
-			os_event_set(buf_flush_event);
-
 			logFreeCheck();
 		}
-
 	}
 
 	/* Convert tuple to rec. */
@@ -1163,7 +1158,7 @@ if no error occurs.
 dberr_t
 BtrBulk::finish(dberr_t	err)
 {
-	ulint		last_page_no = FIL_NULL;
+	uint32_t last_page_no = FIL_NULL;
 
 	ut_ad(!m_index->table->is_temporary());
 
