@@ -39,7 +39,7 @@
 #pragma implementation                          // gcc: Class implementation
 #endif
 
-#include <my_config.h>
+#include <my_global.h>
 #define MYSQL_SERVER 1                          // to have THD
 /* For the moment, include code to deal with integer latches.
  * I have wrapped it with this #ifdef to make it easier to find and remove in the future.
@@ -87,6 +87,7 @@ static const oqgraph_latch_op_table latch_ops_table[] = {
   { "", oqgraph::NO_SEARCH } ,  // suggested by Arjen, use empty string instead of no_search
   { "dijkstras", oqgraph::DIJKSTRAS } ,
   { "breadth_first", oqgraph::BREADTH_FIRST } ,
+  { "leaves", oqgraph::LEAVES },
   { NULL, -1 }
 };
 
@@ -178,7 +179,6 @@ static int oqgraph_init(void *p)
   handlerton *hton= (handlerton *)p;
   DBUG_PRINT( "oq-debug", ("oqgraph_init"));
 
-  hton->state= SHOW_OPTION_YES;
   hton->db_type= DB_TYPE_AUTOASSIGN;
   hton->create= oqgraph_create_handler;
   hton->flags= HTON_ALTER_NOT_SUPPORTED;
@@ -192,6 +192,7 @@ static int oqgraph_init(void *p)
   hton->discover_table_structure= oqgraph_discover_table_structure;
 
   hton->close_connection = oqgraph_close_connection;
+  hton->drop_table= [](handlerton *, const char*) { return 0; };
 
   oqgraph_init_done= TRUE;
   return 0;
@@ -297,7 +298,7 @@ int ha_oqgraph::oqgraph_check_table_structure (TABLE *table_arg)
 
   Field **field= table_arg->field;
   for (i= 0; *field && skel[i].colname; i++, field++) {
-    DBUG_PRINT( "oq-debug", ("Column %d: name='%s', expected '%s'; type=%d, expected %d.", i, (*field)->field_name, skel[i].colname, (*field)->type(), skel[i].coltype));
+    DBUG_PRINT( "oq-debug", ("Column %d: name='%s', expected '%s'; type=%d, expected %d.", i, (*field)->field_name.str, skel[i].colname, (*field)->type(), skel[i].coltype));
     bool badColumn = false;
     bool isLatchColumn = strcmp(skel[i].colname, "latch")==0;
     bool isStringLatch = true;
@@ -346,7 +347,7 @@ int ha_oqgraph::oqgraph_check_table_structure (TABLE *table_arg)
       push_warning_printf( current_thd, Sql_condition::WARN_LEVEL_WARN, HA_WRONG_CREATE_OPTION, "Column %d must be NULL.", i);
     }
     /* Check the column name */
-    if (!badColumn) if (strcmp(skel[i].colname,(*field)->field_name)) {
+    if (!badColumn) if (strcmp(skel[i].colname,(*field)->field_name.str)) {
       badColumn = true;
       push_warning_printf( current_thd, Sql_condition::WARN_LEVEL_WARN, HA_WRONG_CREATE_OPTION, "Column %d must be named '%s'.", i, skel[i].colname);
     }
@@ -562,11 +563,11 @@ int ha_oqgraph::open(const char *name, int mode, uint test_if_locked)
   init_tmp_table_share( thd, share, table->s->db.str, table->s->db.length, options->table_name, "");
   // because of that, we need to reinitialize the memroot (to reset MY_THREAD_SPECIFIC flag)
   DBUG_ASSERT(share->mem_root.used == NULL); // it's still empty
-  init_sql_alloc(&share->mem_root, TABLE_ALLOC_BLOCK_SIZE, 0, MYF(0));
+  init_sql_alloc(PSI_INSTRUMENT_ME, &share->mem_root, TABLE_ALLOC_BLOCK_SIZE, 0, MYF(0));
 
   // What I think this code is doing:
   // * Our OQGRAPH table is `database_blah/name`
-  // * We point p --> /name (or if table happened to be simply `name`, to `name`, dont know if this is possible)
+  // * We point p --> /name (or if table happened to be simply `name`, to `name`, don't know if this is possible)
   // * plen seems to be then set to length of `database_blah/options_data_table_name`
   // * then we set share->normalized_path.str and share->path.str to `database_blah/options_data_table_name`
   // * I assume that this verbiage is needed so  the memory used by share->path.str is set in the share mem root
@@ -577,11 +578,10 @@ int ha_oqgraph::open(const char *name, int mode, uint test_if_locked)
   size_t tlen= strlen(options->table_name);
   size_t plen= (int)(p - name) + tlen + 1;
 
-  share->path.str= (char*)alloc_root(&share->mem_root, plen + 1); // MDEV-5996 space for trailing zero
-  // it seems there was a misunderstanding of why there is a separate length field in the String object
-  strmov(strnmov(share->path.str, name, (int)(p - name) + 1), options->table_name);
-
-  share->path.str[plen] = 0; // MDEV-5996 Make sure the pointer is zero terminated.  I really think this needs refactoring, soon...
+  share->path.str= (char*)alloc_root(&share->mem_root, plen + 1);
+  strmov(strnmov((char*) share->path.str, name, (int)(p - name) + 1),
+         options->table_name);
+  DBUG_ASSERT(strlen(share->path.str) == plen);
   share->normalized_path.str= share->path.str;
   share->path.length= share->normalized_path.length= plen;
 
@@ -622,7 +622,8 @@ int ha_oqgraph::open(const char *name, int mode, uint test_if_locked)
     DBUG_RETURN(-1);
   }
 
-  if (enum open_frm_error err= open_table_from_share(thd, share, "",
+  if (enum open_frm_error err= open_table_from_share(thd, share,
+                                                     &empty_clex_str,
                             (uint) (HA_OPEN_KEYFILE | HA_TRY_READ_ONLY),
                             EXTRA_RECORD,
                             thd->open_options, edges, FALSE))
@@ -655,7 +656,7 @@ int ha_oqgraph::open(const char *name, int mode, uint test_if_locked)
 
   for (Field **field= edges->field; *field; ++field)
   {
-    if (strcmp(options->origid, (*field)->field_name))
+    if (strcmp(options->origid, (*field)->field_name.str))
       continue;
     if ((*field)->cmp_type() != INT_RESULT ||
         !((*field)->flags & NOT_NULL_FLAG))
@@ -680,7 +681,7 @@ int ha_oqgraph::open(const char *name, int mode, uint test_if_locked)
 
   for (Field **field= edges->field; *field; ++field)
   {
-    if (strcmp(options->destid, (*field)->field_name))
+    if (strcmp(options->destid, (*field)->field_name.str))
       continue;
     if ((*field)->type() != origid->type() ||
         !((*field)->flags & NOT_NULL_FLAG))
@@ -703,7 +704,7 @@ int ha_oqgraph::open(const char *name, int mode, uint test_if_locked)
   }
 
   // Make sure origid column != destid column
-  if (strcmp( origid->field_name, destid->field_name)==0) {
+  if (strcmp( origid->field_name.str, destid->field_name.str)==0) {
     fprint_error("Invalid OQGRAPH backing store ('%s.destid' attribute set to same column as origid attribute)", p+1, options->table_name);
     closefrm(edges);
     free_table_share(share);
@@ -712,7 +713,7 @@ int ha_oqgraph::open(const char *name, int mode, uint test_if_locked)
 
   for (Field **field= edges->field; options->weight && *field; ++field)
   {
-    if (strcmp(options->weight, (*field)->field_name))
+    if (strcmp(options->weight, (*field)->field_name.str))
       continue;
     if ((*field)->result_type() != REAL_RESULT ||
         !((*field)->flags & NOT_NULL_FLAG))
@@ -798,12 +799,12 @@ void ha_oqgraph::update_key_stats()
 }
 
 
-int ha_oqgraph::write_row(byte * buf)
+int ha_oqgraph::write_row(const byte * buf)
 {
   return HA_ERR_TABLE_READONLY;
 }
 
-int ha_oqgraph::update_row(const byte * old, byte * buf)
+int ha_oqgraph::update_row(const uchar * old, const uchar * buf)
 {
   return HA_ERR_TABLE_READONLY;
 }
@@ -1185,8 +1186,10 @@ int ha_oqgraph::rename_table(const char *, const char *)
 }
 
 
-ha_rows ha_oqgraph::records_in_range(uint inx, key_range *min_key,
-                                  key_range *max_key)
+ha_rows ha_oqgraph::records_in_range(uint inx,
+                                     const key_range *min_key,
+                                     const key_range *max_key,
+                                     page_range *pages)
 {
   if (graph->get_thd() != current_thd) {
     DBUG_PRINT( "oq-debug", ("g->table->in_use: 0x%lx <-- current_thd 0x%lx", (long) graph->get_thd(), (long) current_thd));
@@ -1210,7 +1213,7 @@ ha_rows ha_oqgraph::records_in_range(uint inx, key_range *min_key,
       min_key->flag != HA_READ_KEY_EXACT ||
       max_key->flag != HA_READ_AFTER_KEY)
   {
-    if (min_key->length == key->key_part[0].store_length && !key->key_part[0].field->is_null()) /* ensure select * from x where latch is null is consistent with no latch */
+    if (min_key && min_key->length == key->key_part[0].store_length && !key->key_part[0].field->is_null()) /* ensure select * from x where latch is null is consistent with no latch */
     {
       // If latch is not null and equals 0, return # nodes
 
