@@ -1,5 +1,5 @@
 /* Copyright (c) 2002, 2013, Oracle and/or its affiliates.
-   Copyright (c) 2009, 2019, MariaDB Corporation.
+   Copyright (c) 2009, 2020, MariaDB Corporation.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -208,18 +208,6 @@ int my_strnncollsp_simple(CHARSET_INFO * cs, const uchar *a, size_t a_length,
 }
 
 
-static int
-my_strnncollsp_nchars_simple(CHARSET_INFO * cs,
-                             const uchar *a, size_t a_length,
-                             const uchar *b, size_t b_length,
-                             size_t nchars)
-{
-  set_if_smaller(a_length, nchars);
-  set_if_smaller(b_length, nchars);
-  return my_strnncollsp_simple(cs, a, a_length, b, b_length);
-}
-
-
 int my_strnncollsp_simple_nopad(CHARSET_INFO * cs,
                                 const uchar *a, size_t a_length,
                                 const uchar *b, size_t b_length)
@@ -1458,7 +1446,7 @@ my_cset_init_8bit(struct charset_info_st *cs, MY_CHARSET_LOADER *loader)
   cs->caseup_multiply= 1;
   cs->casedn_multiply= 1;
   cs->pad_char= ' ';
-  if (!cs->to_lower || !cs->to_upper || !cs->ctype || !cs->tab_to_uni)
+  if (!cs->to_lower || !cs->to_upper || !cs->m_ctype || !cs->tab_to_uni)
     return TRUE;
   return create_fromuni(cs, loader);
 }
@@ -1508,7 +1496,7 @@ int my_mb_ctype_8bit(CHARSET_INFO *cs, int *ctype,
     *ctype= 0;
     return MY_CS_TOOSMALL;
   }
-  *ctype= cs->ctype[*s + 1];
+  *ctype= cs->m_ctype[*s + 1];
   return 1;
 }
 
@@ -1807,10 +1795,9 @@ ret_sign:
   {
     if (negative)
     {
-      if (ull >= (ulonglong) LONGLONG_MIN)
+      if (ull > (ulonglong) LONGLONG_MIN)
       {
-        if (ull != (ulonglong) LONGLONG_MIN)
-          *error= MY_ERRNO_ERANGE;
+        *error= MY_ERRNO_ERANGE;
         return (ulonglong) LONGLONG_MIN;
       }
       *error= 0;
@@ -2035,14 +2022,14 @@ my_strxfrm_pad_desc_and_reverse(CHARSET_INFO *cs,
   if (nweights && frmend < strend && (flags & MY_STRXFRM_PAD_WITH_SPACE))
   {
     uint fill_length= MY_MIN((uint) (strend - frmend), nweights * cs->mbminlen);
-    cs->cset->fill(cs, (char*) frmend, fill_length, cs->pad_char);
+    my_ci_fill(cs, (char*) frmend, fill_length, cs->pad_char);
     frmend+= fill_length;
   }
   my_strxfrm_desc_and_reverse(str, frmend, flags, level);
   if ((flags & MY_STRXFRM_PAD_TO_MAXLEN) && frmend < strend)
   {
     size_t fill_length= strend - frmend;
-    cs->cset->fill(cs, (char*) frmend, fill_length, cs->pad_char);
+    my_ci_fill(cs, (char*) frmend, fill_length, cs->pad_char);
     frmend= strend;
   }
   return frmend - str;
@@ -2101,6 +2088,7 @@ MY_CHARSET_HANDLER my_charset_8bit_handler=
     my_well_formed_char_length_8bit,
     my_copy_8bit,
     my_wc_mb_bin, /* native_to_mb */
+    my_wc_to_printable_8bit
 };
 
 MY_COLLATION_HANDLER my_collation_8bit_simple_ci_handler =
@@ -2108,7 +2096,6 @@ MY_COLLATION_HANDLER my_collation_8bit_simple_ci_handler =
     my_coll_init_simple,	/* init */
     my_strnncoll_simple,
     my_strnncollsp_simple,
-    my_strnncollsp_nchars_simple,
     my_strnxfrm_simple,
     my_strnxfrmlen_simple,
     my_like_range_simple,
@@ -2125,7 +2112,6 @@ MY_COLLATION_HANDLER my_collation_8bit_simple_nopad_ci_handler =
     my_coll_init_simple,	/* init */
     my_strnncoll_simple,
     my_strnncollsp_simple_nopad,
-    my_strnncollsp_nchars_simple,
     my_strnxfrm_simple_nopad,
     my_strnxfrmlen_simple,
     my_like_range_simple,
